package app;

import constant.EvolutionConstants;
import model.Individual;
import model.Population;
import model.SetupEvolutionDTO;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import service.api.CrossoverService;
import service.api.FitnessFunctionService;
import service.api.SelectionService;
import service.impl.crossover.OnePointCrossover;
import service.impl.fitness.MyFitnessFunction;
import service.impl.selection.RouletteSelection;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

/**
 * Tato trida slouzi pouze pro samostatne spusteni evolucniho algoritmu
 */
public class App {
    private final static Logger logger = Logger.getLogger(App.class);

    private FitnessFunctionService defaultFitnessFunctionService = new MyFitnessFunction();

    private App(){
        logger.info("Evolution algorithm started ...");
        //create helper object with settings of process EA
        SetupEvolutionDTO setupProcess = createSetupProcessDTO();

        //create services
        SelectionService selectionService = (SelectionService) createServiceInstance(setupProcess.getSelectionServiceName());
        CrossoverService crossoverService = (CrossoverService) createServiceInstance(setupProcess.getCrossOverServiceName());
        FitnessFunctionService fitnessFunctionService = (FitnessFunctionService) createServiceInstance(setupProcess.getFitnessServiceName());
        if (fitnessFunctionService != null) {
            defaultFitnessFunctionService = fitnessFunctionService;
        }

        if (selectionService == null) {
            selectionService = new RouletteSelection();
        }

        if (crossoverService == null) {
            crossoverService = new OnePointCrossover();
        }
        EvolutionAPI evolution = new MyEvolution(selectionService, crossoverService);

        //first population
        Population population = evolution.createInitialPopulation(setupProcess);

        for (int generationCount = 0; generationCount < setupProcess.getCountGeneration(); generationCount++) {
            calculateQualityPopulation(population);

            population = evolution.createNextPopulation(population);

            // log it
            String outMessage = "Generation " + generationCount + ", Fitness: " + population.getBestIndividual().getQuality() + " Chromosome: "+population.getBestIndividual().getChromosome();
            App.logger.info(outMessage);
        }
        logger.info("done.");
    }

    private SetupEvolutionDTO createSetupProcessDTO() {
        Properties properties = loadProperties(false);

        SetupEvolutionDTO setupProcess = new SetupEvolutionDTO();
        setupProcess.setCrossRate(Double.parseDouble(properties.getProperty(EvolutionConstants.CROSS_RATE)));
        setupProcess.setMutationRate(Double.parseDouble(properties.getProperty(EvolutionConstants.MUTATION_RATE)));
        setupProcess.setCountGeneration(Integer.parseInt(properties.getProperty(EvolutionConstants.GENERATIONS_NUM)));
        setupProcess.setLengthOfChromosome(Integer.parseInt(properties.getProperty(EvolutionConstants.LENGTH_OF_CHROMOSOME)));
        setupProcess.setCountOfPositiveGens(Integer.parseInt(properties.getProperty(EvolutionConstants.COUNT_OF_POSITIVE_GENS)));
        setupProcess.setPenalty(Boolean.parseBoolean(properties.getProperty(EvolutionConstants.PENALTY)));

        int populationSize = Integer.parseInt(properties.getProperty(EvolutionConstants.POPULATION_SIZE));
        int socialPressure = Integer.parseInt(properties.getProperty(EvolutionConstants.SOCIAL_PRESSURE));
        setupProcess.setSizeOfPopulation(populationSize - socialPressure);
        setupProcess.setSocialPressure(socialPressure);

        setupProcess.setSelectionServiceName(properties.getProperty(EvolutionConstants.SELECTION_SERVICE));
        setupProcess.setCrossOverServiceName(properties.getProperty(EvolutionConstants.CROSSOVER_SERVICE));
        setupProcess.setFitnessServiceName(properties.getProperty(EvolutionConstants.FITNESS_SERVICE));

        return setupProcess;
    }

    private Object createServiceInstance(String implementationName) {
        try {
            Class<?> classDefinition = Class.forName(implementationName);
            return classDefinition.newInstance();
        } catch (ClassNotFoundException | InstantiationException | IllegalAccessException e) {
            logger.warn(String.format("It is not able to create service with name %s. Use default!", implementationName), e);
        }
        return null;
    }

    /**
     * Try to get properties from external file. When does not exist than take from classpath
     */
    private Properties loadProperties(boolean useDefault) {
        Properties properties = new Properties();
        File externalConfigFile = new File(EvolutionConstants.EXTERNAL_CONFIG_FILE_NAME);

        if (!useDefault && externalConfigFile.exists()) {
            try (InputStream input = new FileInputStream(externalConfigFile)) {
                properties.load(input);
                logger.info(String.format("Evolution algorithm use configuration from %s", externalConfigFile.getAbsolutePath()));
                if (!validateExternalProp(properties)) {
                    return loadProperties(true);
                }
            } catch (IOException e) {
                logger.error("Error during read external configuration file, Details: {}", e);
            }
        } else {
            try(InputStream input = getClass().getClassLoader().getResourceAsStream(EvolutionConstants.DEFAULT_CONFIG_FILE_NAME)) {
                properties.load(input);
                logger.info("Evolution algorithm use default configuration");
            } catch (IOException e) {
                logger.error("Error during read default configuration file, Details: {}", e);
            }
        }
        return properties;
    }

    private boolean validateExternalProp(Properties properties) {
        if (properties.size() < 11) {
            logger.warn("Some configuration property is missing. Will be used default configuration.");
            return false;
        }

        for (Object key : properties.keySet()) {
            if (StringUtils.isBlank(properties.getProperty((String) key))) {
                logger.warn("Some configuration property is empty. Will be used default configuration.");
                return false;
            }
        }
        return true;
    }

    private void calculateQualityPopulation(Population population) {

        for (Individual individual : population.getIndividuals()) {
            double newQuality = defaultFitnessFunctionService.calculateFitnessFunction(individual);
            individual.setQuality(newQuality);
        }
    }

    public static void main(String[] args) {
        new App();
    }
}