package app;

import model.Individual;
import model.Population;
import model.SetupEvolutionDTO;
import service.api.CrossoverService;
import service.api.SelectionService;
import service.impl.crossover.OnePointCrossover;
import service.impl.selection.RouletteSelection;
import utils.EvolutionUtils;

import java.util.Set;

/**
 * Created with IntelliJ IDEA.
 * User: lukas
 * Date: 14.5.15
 */
public class MyEvolution implements EvolutionAPI {

    private SelectionService selectionService = new RouletteSelection();
    private CrossoverService crossoverService = new OnePointCrossover();

    public MyEvolution() {
    }

    public MyEvolution(SelectionService selection, CrossoverService crossover) {
        this.selectionService = selection;
        this.crossoverService = crossover;
    }

    @Override
    public Population createInitialPopulation(SetupEvolutionDTO setupProcess) {
        // create init Population
        Population initPopulation = new Population(setupProcess);
        // create individuals with random chromosome
        initPopulation.initPopulation();
        return initPopulation;
    }

    @Override
    public Population createNextPopulation(Population previousPopulation) {
        //urcite nejlepsiho a nejhorsiho z populace
        preProcessNextPopulation(previousPopulation);

        // upravime hodnoty fitness pro lepsi selekci od 0 do bestFitness-minFitness
        double sumFitness = 0.0;
        for (Individual individual : previousPopulation.getIndividuals()) {
            double quality= individual.getQuality() - previousPopulation.getWorstIndividual().getQuality();
            individual.setQuality(quality);
            sumFitness += quality;
        }
        // create better population than before and replace previously population
        return createNewPopulation(previousPopulation, sumFitness);
    }

    /**
     * Firstly resolve if individual has penalty and then finds the best and worst individuals of the population.
     */
    private void preProcessNextPopulation(Population population) {
        double bestFitness = 0.0f;
        double minFitness = Double.MAX_VALUE;

        for (Individual individual : population.getIndividuals()) {
            //pokud jedinec ma vice jednicek nez bylo definovano -> trest
            if (population.isPenalty() && doPenalty(individual, population.getCountOfPositiveGens())) {
                continue;
            }
            //jinak hledame nejlepsiho / nejhorsiho
            double newQuality = individual.getQuality();

            if (newQuality > bestFitness) {
                bestFitness = newQuality;
                population.setBestIndividual(individual);
            }
            if (newQuality < minFitness) {
                minFitness = newQuality;
                population.setWorstIndividual(individual);
            }
        }

    }

    /**
     * When individual has more than define count of positive gens then must be penalty.
     */
    private boolean doPenalty(Individual individual, Integer countOfPositiveGens) {
        if (countOfPositiveGens < EvolutionUtils.getCountOfPositiveGens(individual.getChromosome())) {
            individual.setQuality(individual.getQuality() / 2);
            return true;
        }
        return false;
    }

    /**
     * Creates better population. Makes selection, crossover and mutation.
     *
     * @param oldPopulation
     *            - previous population
     * @param sumFitness
     *            - SUM of all fitness of old population
     * @return better population
     */
    private Population createNewPopulation(Population oldPopulation, double sumFitness) {
        Population newPopulation = new Population(oldPopulation);
        // iterate over individuals of population
        for (int iIndividual = 0; iIndividual < oldPopulation.getSize(); iIndividual += 2) {
            // random select two individuals
            Individual individual1 = selectionService.randomSelectIndividual(oldPopulation.getIndividuals(), sumFitness);
            Individual individual2 = selectionService.randomSelectIndividual(oldPopulation.getIndividuals(), sumFitness);
            // crossover - create offspring
            if (Math.random() < oldPopulation.getCrossRate()) {
                Individual[] children = crossoverService.crossover(individual1, individual2);
                newPopulation.getIndividuals().add(iIndividual, children[0]);
                newPopulation.getIndividuals().add(iIndividual + 1, children[1]);

                // mutation
                for (Individual child : children) {
                    if (Math.random() < newPopulation.getMutationRate()) {
                        doMutation(child);
                    }
                }
            } else { // if not crossover then ..
                newPopulation.getIndividuals().add(iIndividual, individual1);
                newPopulation.getIndividuals().add(iIndividual + 1, individual2);
            }
        }
        // the best individual must live
        newPopulation.getIndividuals().set(0, oldPopulation.getBestIndividual());
        newPopulation.setBestIndividual(oldPopulation.getBestIndividual());

        //do it social injection
        doItSocialInjection(newPopulation);
        return newPopulation;
    }

    /**
     * Put new random individual into exist population. Social injection
     */
    private void doItSocialInjection(Population population) {
        for (int i = 0; i < population.getSocialPressure(); i++) {
            Individual randomIndividual = population.createRandomIndividual();
            population.getIndividuals().add(randomIndividual);
        }
    }

    private void doMutation(Individual child) {
        int length = child.getChromosome().length();
        String originalChromosome = child.getChromosome();
        //get random positions where will be mutation
        Set<Integer> generatedPositionToChange = child.getRandomPositions(length, EvolutionAPI.randomGenerator.nextInt(length));
        //mutation on random positions
        for (Integer position : generatedPositionToChange) {
            char gen = originalChromosome.charAt(position);
            if (gen == '1') {
                child.setChromosome(EvolutionUtils.changeCharInPosition(position, '0', originalChromosome));
            } else {
                child.setChromosome(EvolutionUtils.changeCharInPosition(position, '1', originalChromosome));
            }
        }
    }


}
